# Copyright (C) 2019 Alteryx, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Base classes for plugin, input/output anchors/managers."""

# Built in Libraries
import copy
import os
import sys
from collections import OrderedDict, namedtuple, UserDict
from functools import partial
from types import SimpleNamespace
from typing import Any, List, Tuple, Union

import AlteryxPythonSDK as sdk

import snakeplane.interface_utilities as interface_utils
import snakeplane.plugin_utilities as plugin_utils
from snakeplane.constants import SNAKEPLANE_NULL_VALUE_PLACEHOLDER

import xmltodict


class AyxPlugin:
    """Base plugin class to be modified by snakeplane."""

    def __init__(
        self,
        n_tool_id: int,
        alteryx_engine: sdk.AlteryxEngine,
        output_anchor_mgr: sdk.OutputAnchorManager,
    ) -> None:
        # Initialization data
        self._engine_vars = SimpleNamespace()
        self._engine_vars.n_tool_id = n_tool_id
        self._engine_vars.alteryx_engine = alteryx_engine
        self._engine_vars.output_anchor_mgr = output_anchor_mgr
        self._raised_missing = False

        # Plugin State vars
        self._state_vars = SimpleNamespace(
            initialized=False,
            input_anchors={},
            output_anchors={},
            config_data=None,
            required_input_names=[],
        )

        # Pull in the config XML data from conf file using the name of the tool
        xml_files = [
            file
            for file in os.listdir(plugin_utils.get_tool_path(self.tool_name))
            if file.lower().endswith(".xml")
        ]
        with open(
            os.path.join(plugin_utils.get_tool_path(self.tool_name), xml_files[0])
        ) as fd:
            self._state_vars.config_data = xmltodict.parse(fd.read())

        # Plugin Error Methods
        self.logging = SimpleNamespace(
            display_error_msg=partial(
                self._engine_vars.alteryx_engine.output_message,
                self._engine_vars.n_tool_id,
                sdk.EngineMessageType.error,
            ),
            display_warn_msg=partial(
                self._engine_vars.alteryx_engine.output_message,
                self._engine_vars.n_tool_id,
                sdk.EngineMessageType.warning,
            ),
            display_info_msg=partial(
                self._engine_vars.alteryx_engine.output_message,
                self._engine_vars.n_tool_id,
                sdk.EngineMessageType.info,
            ),
        )

        # Default to no inputs or outputs
        for connection in plugin_utils.get_xml_config_input_connections(
            self._state_vars.config_data
        ):
            self._state_vars.input_anchors[connection["@Name"]] = []

            # Track names of the inputs that are required for this tool to run
            if connection["@Optional"] == "False":
                self._state_vars.required_input_names.append(connection["@Name"])

        for connection in plugin_utils.get_xml_config_output_connections(
            self._state_vars.config_data
        ):
            self._state_vars.output_anchors[connection["@Name"]] = OutputAnchor()

        # Custom data
        self.user_data = SimpleNamespace()

        # Configure managers, this must occur last so the instance
        # is properly configured
        self.input_manager = InputManager(self)
        self.output_manager = OutputManager(self)

    @property
    def initialized(self) -> bool:
        """Getter for plugin initialization state."""
        return self._state_vars.initialized

    @initialized.setter
    def initialized(self, value: bool) -> None:
        """Setter for plugin initialization state."""
        self._state_vars.initialized = bool(value)

    @property
    def update_only_mode(self) -> bool:
        """Getter for if designer is in update only mode."""
        return (
            self._engine_vars.alteryx_engine.get_init_var(
                self._engine_vars.n_tool_id, "UpdateOnly"
            )
            == "True"
        )

    @property
    def all_inputs_completed(self) -> bool:
        """
        Check that all required inputs have successfully completed.

        Parameters
        ----------
        current_plugin : object
            An AyxPlugin object

        Returns
        -------
        bool
            Boolean indication of if all inputs have completed.
        """
        all_inputs_completed = True
        if self.initialized:
            for name in self._state_vars.required_input_names:
                connections = self._state_vars.input_anchors[name]
                if len(connections) == 0 or not all(
                    [connection.completed for connection in connections]
                ):
                    all_inputs_completed = False
        else:
            all_inputs_completed = False
        return all_inputs_completed

    @property
    def all_required_inputs_initialized(self) -> bool:
        """Getter for checking if all required inputs have been initialized."""
        for anchor_name in self._state_vars.required_input_names:
            input = self._state_vars.input_anchors[anchor_name]
            if not input or not all([connection.initialized for connection in input]):
                return False

        return True

    def update_sys_path(self) -> None:
        """Update the sys path to include the current tools libs."""
        # Add lib to sys path
        tool_path = plugin_utils.get_tool_path(self.tool_name)
        sys.path.append(tool_path)
        sys.path.append(os.path.join(tool_path, "Lib", "site-packages"))

    def assert_all_inputs_connected(self) -> bool:
        """Raise an error if there are any missing input connections."""
        for anchor_name in self._state_vars.required_input_names:
            input = self._state_vars.input_anchors[anchor_name]
            if not input:
                if not self._raised_missing:
                    self.logging.display_error_msg("Missing Incoming Connection(s).")
                    self._raised_missing = True
                return False

        return True

    def save_output_anchor_refs(self) -> None:
        """Save all references to output anchors."""
        # Get references to the output anchors
        for anchor_name in self._state_vars.output_anchors:
            self._state_vars.output_anchors[
                anchor_name
            ]._handler = self._engine_vars.output_anchor_mgr.get_output_anchor(
                anchor_name
            )

    def save_interface(self, name: str, interface: object) -> None:
        """Save the interface internally."""
        self._state_vars.input_anchors[name].append(interface)

    def update_progress(self, d_percentage: float) -> None:
        """Update the progress on this anchor."""
        self._engine_vars.alteryx_engine.output_tool_progress(
            self._engine_vars.n_tool_id, d_percentage
        )  # Inform the Alteryx engine of the tool's progress.

        for _, anchor in self._state_vars.output_anchors.items():
            # Inform the downstream tool of this tool's progress.
            anchor._handler.update_progress(d_percentage)

    def close_all_outputs(self) -> None:
        """Force all output anchors to close."""
        # Close all output anchors
        for _, anchor in self._state_vars.output_anchors.items():
            anchor._handler.close()

        # Checks whether connections were properly closed.
        for anchor_name in self._state_vars.output_anchors:
            self._state_vars.output_anchors[anchor_name]._handler.assert_close()

    def push_all_output_records(self) -> None:
        """
        For each output anchor on the plugin, flush all the output records.

        Parameters
        ----------
        current_plugin: object
            The plugin for which to flush output records

        Returns
        -------
        None
        """
        for _, output_anchor in self._state_vars.output_anchors.items():
            output_anchor.push_records(self)

    def push_all_metadata(self) -> None:
        """Pushes all output anchor metadata downstream."""
        for _, anchor in self._state_vars.output_anchors.items():
            anchor.push_metadata(self)

    def clear_accumulated_records(self) -> None:
        """
        Clear all accumulated records from all plugin interfaces.

        Parameters
        ----------
        plugin: object
            The plugin to clear all records from

        Returns
        -------
        None
            This function has side effects on plugin, and therefore has no return
        """
        for _, anchor in self._state_vars.input_anchors.items():
            for connection in anchor:
                connection._interface_record_vars.record_list_in = []

    def create_record_info(self) -> sdk.RecordInfo:
        """Create a new record info object."""
        return sdk.RecordInfo(self._engine_vars.alteryx_engine)


class AyxPluginInterface:
    """Input interface base definition."""

    def __init__(self, parent: object, name: str) -> None:
        self.parent = parent
        self.name = name
        self.initialized = False

        self._interface_record_vars = SimpleNamespace(
            record_info_in=None, record_list_in=[], column_metadata=None
        )

        self._interface_state = SimpleNamespace(
            input_complete=False, d_progress_percentage=0, data_processing_mode="batch"
        )

        self.is_last_chunk = None

    @property
    def metadata(self) -> object:
        """Input metadata getter."""
        return copy.deepcopy(self._interface_record_vars.column_metadata)

    @property
    def data(self) -> Union[object, List[List[Any]]]:
        """Input data getter."""
        if (
            self.parent.process_data_mode == "stream"
            and self.parent.process_data_input_type == "list"
        ):
            return self._interface_record_vars.record_list_in[0]
        elif self.parent.process_data_input_type == "list":
            return self._interface_record_vars.record_list_in
        else:
            try:
                import pandas as pd
            except ImportError:
                err_str = """The Pandas library must be installed to
                            allow dataframe as input_type."""
                self.parent.logging.display_error_msg(err_str)
                raise ImportError(err_str)
            else:
                return pd.DataFrame(
                    self._interface_record_vars.record_list_in,
                    columns=self.metadata.get_column_names(),
                )

    @property
    def completed(self) -> bool:
        """Interface completed getter."""
        return self._interface_state.input_complete

    @completed.setter
    def completed(self, val) -> None:
        """Interface completed setter."""
        self._interface_state.input_complete = val

    @property
    def anchor_metadata(self):
        """Anchor metadata getter."""
        return self._interface_record_vars.column_metadata

    @anchor_metadata.setter
    def anchor_metadata(self, val) -> None:
        """Anchor metadata setter."""
        self._interface_record_vars.column_metadata = val

    @property
    def record_info(self):
        """Getter for Input Anchor record_info object."""
        return self._interface_record_vars.record_info_in

    def get_values_from_record(
        self: object, in_record: object
    ) -> Tuple[List[Union[int, float, bool, str, bytes]], dict]:
        """
        Get a list of values from an incoming record.

        Parameters
        ----------
        interface_obj : object
            An AyxPluginInterface object for the current interface

        in_record: object
            An Alteryx RecordRef object for the record to be processed
        Returns
        ---------
        Tuple[List[int, float, bool, str, bytes], dict]
            The return takes the form (record, metadata)
            where:
                record: A list of the parsed record values
                metadata: a dict containing the names, types, sizes,
                sources, and descriptions of each field
        """
        fields = self._interface_record_vars.fields
        field_getters = self._interface_record_vars.field_getters

        row = [field_getters[field](in_record) for field in fields]
        return row

    def accumulate_record(self, record: sdk.RecordRef) -> None:
        """Accumulate an incoming record."""
        row = self.get_values_from_record(record)

        self._interface_record_vars.record_list_in.append(row)


class InputManager(UserDict):
    """Manager of input anchors with helper functions."""

    def __init__(self, plugin: object) -> None:
        self._plugin = plugin
        self.data = self._plugin._state_vars.input_anchors

    @property
    def tool_id(self) -> int:
        """Getter for the current tool ID."""
        return self._plugin._engine_vars.n_tool_id

    @property
    def workflow_config(self) -> OrderedDict:
        """Getter for the workflow config."""
        return self._plugin.workflow_config


class OutputManager(UserDict):
    """Manager of output anchors."""

    def __init__(self, plugin: object) -> None:
        self._plugin = plugin
        self.data = self._plugin._state_vars.output_anchors

    def get_temp_file_path(self) -> str:
        """Create a temp file using the Alteryx Engine."""
        return self._plugin._engine_vars.alteryx_engine.create_temp_file_name()

    @staticmethod
    def create_anchor_metadata():
        """Create a new anchor metadata object."""
        return AnchorMetadata()


class OutputAnchor:
    """Output anchor bookkeeping class with helpers."""

    def __init__(self) -> None:
        self._data = None
        self._metadata = None
        self._record_info_out = None
        self._record_creator = None
        self._handler = None

    @property
    def data(self) -> Union[object, List[List[Any]]]:
        """Getter for anchor data."""
        return self._data

    @data.setter
    def data(self, data: Union[object, List[List[Any]]]) -> None:
        """Setter for anchor data."""
        self._data = data

    @property
    def metadata(self) -> object:
        """Getter for the anchor metadata."""
        return copy.deepcopy(self._metadata)

    @metadata.setter
    def metadata(self, metadata):
        """Setter for anchor metadata."""
        self._metadata = metadata

    def get_data_list(self) -> List[List[Any]]:
        """Get the list of data to push downstream as a list of lists."""
        if interface_utils.is_dataframe(self._data):
            return interface_utils.dataframe_to_list(self._data)
        elif type(self._data) == list and not type(self._data[0]) == list:
            return [self._data]
        return self._data

    def push_metadata(self: object, plugin: object) -> None:
        """Propagate the metadata downstream for this anchor."""
        out_col_metadata = self.metadata
        if out_col_metadata is None:
            return

        if self._record_info_out is None:

            self._record_info_out = plugin.create_record_info()

            interface_utils.build_ayx_record_info(
                out_col_metadata, self._record_info_out
            )

            self._handler.init(self._record_info_out)

    def push_records(self, plugin: object) -> None:
        """
        Flush all records for an output anchor.

        Parameters
        ----------
        current_plugin: object
            The plugin that the output belongs to

        output_anchor: object
            The output anchor to flush

        Returns
        -------
        None
        """

        out_values_list = self.get_data_list()
        out_col_metadata = self.metadata

        # If there are no output records, just return
        if out_values_list is None:
            return

        if not self._record_info_out:
            self.push_metadata(plugin)

        Column = namedtuple(
            "Column", ["name", "type", "size", "scale", "source", "description"]
        )

        columns = [
            Column(
                out_col_metadata[i].name,
                out_col_metadata[i].type,
                out_col_metadata[i].size,
                out_col_metadata[i].scale,
                out_col_metadata[i].source,
                out_col_metadata[i].description,
            )
            for i in range(len(out_col_metadata))
        ]

        name_to_field_dict = {
            column.name: self._record_info_out.get_field_by_name(column.name)
            for column in columns
        }

        name_to_setter_dict = {
            column.name: interface_utils.get_field_setter_from_type(
                name_to_field_dict[column.name]
            )
            for column in columns
        }

        record_creator = self._record_info_out.construct_record_creator()

        for row in out_values_list:
            record_creator.reset()

            for col_idx, column in enumerate(columns):
                field = name_to_field_dict[column.name]
                element = row[col_idx]
                if element is None or element == SNAKEPLANE_NULL_VALUE_PLACEHOLDER:
                    field.set_null(record_creator)
                else:
                    name_to_setter_dict[column.name](record_creator, element)

            ayx_record = record_creator.finalize_record()

            self._handler.push_record(ayx_record, False)

        # Clear the data from the output_anchor
        self.data = None


class ColumnMetadata:
    """Column Metadata tracking class."""

    def __init__(
        self,
        name: str,
        col_type: sdk.FieldType,
        size: int,
        scale: int,
        source: str,
        description: str,
    ):
        self.name = name
        self.type = col_type
        self.size = size
        self.scale = scale
        self.source = source
        self.description = description

    def __deepcopy__(self, memo):
        """Override of deep copy method."""
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k not in ["type"]:
                setattr(result, k, copy.deepcopy(v, memo))

        setattr(result, "type", self.type)
        return result

    def __iter__(self):
        """Generate the iterable for this class."""
        for el in [
            self.name,
            self.type,
            self.size,
            self.scale,
            self.source,
            self.description,
        ]:
            yield el


class AnchorMetadata:
    """Class for tracking column metadata for a given anchor."""

    def __init__(self):
        self.columns = []

    @property
    def columns(self) -> List[ColumnMetadata]:
        """Getter for columns."""
        return self._columns

    @columns.setter
    def columns(self, value: List[ColumnMetadata]):
        """Setter for columns."""
        self._columns = value

    def add_column(
        self,
        name: str,
        col_type: sdk.FieldType,
        size: int = 256,
        scale: int = 0,
        source: str = "",
        description: str = "",
    ):
        """Add a column to this anchor."""
        self.columns.append(
            ColumnMetadata(name, col_type, size, scale, source, description)
        )

    def index_of(self, name: str):
        """Get the column index of a given column name."""
        try:
            return [c.name for c in self.columns].index(name)
        except ValueError:
            return None

    def get_column_by_name(self, name: str) -> ColumnMetadata:
        """Get the column given the column name."""
        index = self.index_of(name)
        if index is None:
            return None
        return self.columns[index]

    def get_column_names(self) -> List[str]:
        """Get a list of the column names available."""
        return [c.name for c in self.columns]

    def __getitem__(self, key: int) -> ColumnMetadata:
        """Get the column specified by key (an index)."""
        return self.columns[key]

    def __len__(self) -> int:
        """Return the number of columns as the length."""
        return len(self.columns)



# Copyright (C) 2019 Alteryx, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Interface utilities for the snakeplane library."""
from snakeplane.constants import SNAKEPLANE_NULL_VALUE_PLACEHOLDER

from typing import Any, List, Optional, Tuple

import AlteryxPythonSDK as sdk


type_dict = {
    "blob": "get_as_blob",
    "byte": "get_as_int32",
    "int16": "get_as_int32",
    "int32": "get_as_int32",
    "int64": "get_as_int64",
    "float": "get_as_double",
    "double": "get_as_double",
    "date": "get_as_string",
    "time": "get_as_string",
    "datetime": "get_as_string",
    "bool": "get_as_bool",
    "string": "get_as_string",
    "v_string": "get_as_string",
    "v_wstring": "get_as_string",
    "wstring": "get_as_string",
    "fixeddecimal": "get_as_double",
    "spatialobj": "get_as_blob",
}


def get_getter_from_field(field):
    return getattr(field, type_dict[str(field.type)])


def get_dynamic_type_value(field: sdk.Field, record: sdk.RecordRef) -> Any:
    """
    Extract a field from a record, dynamically calculating its type.

    Parameters
    ----------
    field : object
        An Alteryx Field object that is present in an Alteryx RecordInfo object.
        Alteryx Field objects contain various attributes, including type, as well
        as the actual methods that allow for getting/setting values in the
        RecordRef passed in via C++ engine.

    record : object
        A single record object from Alteryx that contains a row of data.

    Returns
    -------
    Any
        The return value of this function can be any of types blob, int32, int64,
        dobule, bool, or string. The returned value represents the parsed/typed
        value of the desired field from the input record
    """
    try:
        return getattr(field, type_dict[str(field.type)])(record)
    except KeyError:
        # The type wasn't found, throw an error to let the use know
        err_str = f"""Failed to automatically convert field type "{str(field.type)}"
                    due to unidentified type name. This is due to a currently unsupported type."""
        raise TypeError(err_str)


def get_column_names_list(record_info_in: sdk.RecordInfo) -> List[str]:
    """
    Extract the column names from an Alteryx record info object.

    Parameters
    ----------
    record_info_in : object
        An Alteryx RecordInfo object
    Returns
    ---------
    List[str]
        A list of the column names in string format
    """
    return [field.name for field in record_info_in]


def get_column_metadata(record_info_in: sdk.RecordInfo) -> dict:
    """
    Extract record metadata from an Alteryx record info object.

    Parameters
    ----------
    record_info_in : object
        An Alteryx RecordInfo object

    Returns
    -------
    List[dict]
        A list of column metadata
    """
    from snakeplane.helper_classes import AnchorMetadata

    metadata = AnchorMetadata()

    for field in record_info_in:
        metadata.add_column(
            field.name,
            field.type,
            size=field.size,
            source=field.source,
            scale=field.scale,
            description=field.description,
        )

    return metadata


def get_column_types_list(record_info_in: sdk.RecordInfo) -> List[object]:
    """
    Collect the column types from an Alteryx record info object.

    Parameters
    ----------
    record_info_in : object
        An Alteryx RecordInfo object
    Returns
    -------
    List[object]
        A list of the column types as per the AlteryxSDK
    """
    return [field.type for field in record_info_in]


type_cast_dict = {
    sdk.FieldType.bool: ("set_from_bool", bool),
    sdk.FieldType.blob: ("set_from_blob", bytes),
    sdk.FieldType.double: ("set_from_double", float),
    sdk.FieldType.float: ("set_from_double", float),
    sdk.FieldType.fixeddecimal: ("set_from_double", float),
    sdk.FieldType.byte: ("set_from_int32", int),
    sdk.FieldType.int16: ("set_from_int32", int),
    sdk.FieldType.int32: ("set_from_int32", int),
    sdk.FieldType.int64: ("set_from_int64", int),
    sdk.FieldType.string: ("set_from_string", str),
    sdk.FieldType.v_string: ("set_from_string", str),
    sdk.FieldType.v_wstring: ("set_from_string", str),
    sdk.FieldType.wstring: ("set_from_string", str),
    sdk.FieldType.date: ("set_from_string", str),
    sdk.FieldType.datetime: ("set_from_string", str),
    sdk.FieldType.time: ("set_from_string", str),
}


def get_field_setter_from_type(field):
    set_attr_name, caster = type_cast_dict[field.type]
    setter = getattr(field, set_attr_name)

    def setter_func(record_creator, value):
        setter(record_creator, caster(value))

    return setter_func


def add_new_field_to_record_info(
    record_info: object,
    field_name: str,
    field_type: object,
    field_size: int,
    field_scale: int,
    field_source: str,
    field_desc: str,
) -> None:
    """
    Attaches a field to the specified Alteryx RecordInfo object.

    Parameters
    ----------
    record_info : object
        An Alteryx RecordInfo object.
        Alteryx RecordInfo objects act as containers for the necessary metadata
        needed by the Alteryx engine to generate, identify, and manipulate
        each record of data passing through the tool.

    field_name : str
        A string representing the desired name of the field.

    field_type : object
        An Alteryx FieldType object.  It can be one of the following:
            sdk.FieldType.string,
            sdk.FieldType.v_string,
            sdk.FieldType.v_wstring,
            sdk.FieldType.wstring,
            sdk.FieldType.bool,
            sdk.FieldType.blob,
            sdk.FieldType.byte,
            sdk.FieldType.int16,
            sdk.FieldType.int32,
            sdk.FieldType.int64,
            sdk.FieldType.float,
            sdk.FieldType.double,
            sdk.FieldType.date,
            sdk.FieldType.datetime,
            sdk.FieldType.time,
            sdk.FieldType.spatial

    field_size : int
        An integer specifying the size of the desired Alteryx Field. This
        option is ignored for primitive types, and is only used for string,
        blob, and spatial types.

    field_source : str
        Where this field came from

    field_desc
        A short description of what this field is

    Returns
    -------
    None
        This is a stateful function that produces side effects by modifying
        the record_info object.
    """
    string_field_set = {
        sdk.FieldType.string,
        sdk.FieldType.v_string,
        sdk.FieldType.v_wstring,
        sdk.FieldType.wstring,
    }
    default_string_size = 255
    if (field_size is None) and (field_type in string_field_set):
        field_size = default_string_size
    elif field_size is None:
        field_size = 0

    record_info.add_field(
        field_name,
        field_type,
        size=field_size,
        scale=field_scale,
        source=field_source,
        description=field_desc,
    )


def build_ayx_record_info(metadata: dict, record_info: sdk.RecordInfo) -> None:
    """
    Create a record info object from a metadata dictionary.

    Populates a an Alteryx RecordInfo object with field objects based on the
    contents of the names and types specified in names_list and types_list,
    respectively.

    Parameters
    ----------
    metadata : dict
        A dict containing all of the names, types, sizes, sources,
        and descriptions of each field. These are used to generate
        the Alteryx RecordInfo object (if it doesn't already exist)
        for the names of each respective Field object.

    record_info : object
        An Alteryx RecordInfo object.
        Alteryx RecordInfo objects act as containers for the necessary metadata
        needed by the Alteryx engine to generate, identify, and manipulate
        each record of data passing through the tool.

    Returns
    -------
    None
        This is a stateful function that produces side effects by modifying
        the record_info object.

    """
    output_columns = metadata.columns

    for output_column in output_columns:
        add_output_column_to_record_info(output_column, record_info)


def add_output_column_to_record_info(
    output_column: tuple, record_info_out: sdk.RecordInfo
) -> None:
    """
    Add a column to a RecordInfo object.

    Parameters
    ----------
    output_column: tuple
        A tuple containing the metadata on the new column (name, type)

    record_info_out: object
        RecordInfo object to append the new column to

    Returns
    -------
    None
    """
    add_new_field_to_record_info(
        record_info_out,
        output_column.name,
        output_column.type,
        output_column.size,
        output_column.scale,
        output_column.source,
        output_column.description,
    )


def is_dataframe(input: Any) -> bool:
    """
    Check if the input variable is a pandas dataframe.

    Parameters
    ----------
    input
        Any variable

    Returns
    -------
    bool
        Indication if the input is a pandas dataframe
    """
    try:
        import pandas as pd
    except ImportError:
        err_str = """The Pandas library must be installed to
                    allow dataframe as input_type."""
        raise ImportError(err_str)
    else:
        return isinstance(input, pd.DataFrame)


def dataframe_to_list(df: object) -> List[List[Any]]:
    """
    Convert a pandas dataframe to a list of lists.

    Parameters
    ----------
    df: object
        A pandas dataframe

    Returns
    -------
    List[List[Any]]
        A list of lists of the pandas dataframe data
    """
    import numpy as np

    df.fillna(SNAKEPLANE_NULL_VALUE_PLACEHOLDER, inplace=True)
    return df.values.tolist()



# Copyright (C) 2019 Alteryx, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Implementation of a plugin factory for use in snakeplane framework."""

# Built in Libraries
import logging
from functools import wraps

# 3rd Party Libraries
import AlteryxPythonSDK as sdk

import snakeplane.interface_utilities as interface_utils
from snakeplane.helper_classes import AyxPlugin, AyxPluginInterface

import xmltodict

DEBUG = False

logger = logging.getLogger(__name__)


class PluginFactory:
    """
    Class for generating a plugin using the snakeplane framework.

    The PluginFactory follows a Flask-like paradigm of leveraging decorators to inject
    custom user methods while abstracting the boilerplate operations for input and
    output with the Alteryx Engine.

    Attributes
    ----------
    plugin : object
        The plugin property contains a reference to a dynamic class declaration for the
        AyxPlugin class required by the Alteryx Engine for a Python SDK Plugin tool.
        The plugin object gets updated and further defined by the user as they call
        methods on the PluginFactory instance (either directly or via decorators).

    plugin.plugin_interface : object
        Although not a direct attribute of the PluginFactory class, due to the
        metaprogramming nature of the PluginFactory, the plugin_interface object is
        also a reference to a dynamic class declaration for an Alteryx Plugin Interface
        that is used as a child class instance by the plugin object. The PluginFactory
        is also dynamically constructing this class declaration based on the user's use
        of the PluginFactory's methods (either directly or via decorators).
    """

    def __init__(self, tool_name: str) -> None:
        """
        Initialize a PluginFactory object.

        Parameters
        ----------
        tool_name : str
        The name of the tool you are generating Python SDK engine code for.
        It is important to note that this name must match the name of the tool
        specified in the name of its Config/Icon files.  For example, a tool called
        ExampleStream would have a config file called "ExampleStreamConfig.xml".
        PluginFactory uses this to find and read in the contents of the xml file
        to better automate input and output creation.

        Examples
        --------
            # Assuming that tool has ExampleToolConfig.xml file
            factory = PluginFactory("ExampleTool")
        """

        # Make local, per instance copies of the plugins so that multiple plugins
        # can be generated with the same library. Since the plugin factory does
        # metaprogramming, we can't modify the original definitions of
        # AyxPlugin/AyxPluginInterface without contaminating the package
        class Plugin(AyxPlugin):
            pass

        class Interface(AyxPluginInterface):
            pass

        self._plugin = Plugin
        self._plugin.plugin_interface = Interface

        setattr(self._plugin, "tool_name", tool_name)

        # Initialize all required methods with default behavior
        def noop(*args, **kwargs) -> None:
            pass

        def always_true(*args, **kwargs):
            return True

        self.build_pi_init(always_true),
        self.build_pi_add_incoming_connection(noop),
        self.build_pi_push_all_records(noop),
        self.build_pi_add_outgoing_connection(always_true)
        self.build_pi_close(noop)

        self.build_ii_init(always_true),
        self.build_ii_push_record(always_true),
        self.build_ii_update_progress(noop),
        self.build_ii_close(noop)
        self.build_metadata(noop)

        self._init_func = always_true

    def build_pi_init(self, func: object):
        """
        Register a custom pi_init method.

        Parameters
        ----------
        func: Callable[object, str]
        The user-defined function that will be called by the Alteryx Engine to
        initialize the plugin.

        Returns
        -------
        None
        This method produces side-effects by registering the user defined function
        """

        @_monitor("pi_init")
        @wraps(func)
        def wrap_pi_init(current_plugin, config_xml):
            try:
                current_plugin.update_sys_path()
                current_plugin.save_output_anchor_refs()

                # Parse XML and save
                current_plugin.workflow_config = xmltodict.parse(
                    config_xml, strip_whitespace=False
                )["Configuration"]

                # Call decorated function
                val = func(current_plugin)

                if (
                    current_plugin.update_only_mode
                    and len(current_plugin._state_vars.required_input_names) == 0
                ):
                    self._init_func(current_plugin)

                # Boilerplate Side Effects
                current_plugin.initialized = val

                return val
            except Exception as e:
                logger.exception(e)
                raise e

        setattr(self._plugin, "pi_init", wrap_pi_init)

    def build_pi_add_incoming_connection(self, func: object):
        """
        Register a custom pi_add_incoming_connection method.

        Parameters
        ----------
        func: Callable[object, str, str]
        The user-defined function that will be called by the Alteryx Engine for each
        incoming connection.  It is expected that this function returns an initialized
        AyxInterface object.

        Returns
        -------
        None
        This method produces side-effects by registering the user defined function
        """

        @_monitor("pi_add_incoming_connection")
        @wraps(func)
        def wrap_pi_add_incoming_connection(
            current_plugin: object, str_type: str, str_name: str
        ):
            # Call decorated function
            func(current_plugin, str_type, str_name)

            # Generate the interface
            interface = current_plugin.plugin_interface(current_plugin, str_type)

            # Save it to the plugin
            current_plugin.save_interface(str_type, interface)

            # Return it
            return interface

        setattr(
            self._plugin, "pi_add_incoming_connection", wrap_pi_add_incoming_connection
        )

    def build_pi_push_all_records(self, func: object):
        """
        Register a custom pi_push_all_records method.

        Parameters
        ----------
        func: Callable[object, int]
        The user-defined function that will be called by the Alteryx Engine when
        no incoming interface exists.  Typically this is for generating records from
        sources outside of Alteryx, such as an API or database.
        It is expected that this function returns a True if no errors are present,
        otherwise False.

        Returns
        -------
        None
        This method produces side-effects by registering the user defined function
        """

        @_monitor("push_all_records")
        @wraps(func)
        def wrap_push_all_records(current_plugin: object, n_record_limit: int):
            try:
                if len(current_plugin._state_vars.required_input_names) == 0:
                    if current_plugin.update_only_mode:
                        self._build_metadata(current_plugin)
                        for (
                            _,
                            anchor,
                        ) in current_plugin._state_vars.output_anchors.items():
                            anchor.push_metadata(current_plugin)
                    else:
                        # Only call the users defined function when there are no required
                        # inputs, since this is the only scenario where something interesting
                        # happens in this function
                        func(current_plugin, n_record_limit)
                else:
                    current_plugin.assert_all_inputs_connected()

                return True

            except Exception as e:
                logger.exception(e)
                raise e

        setattr(self._plugin, "pi_push_all_records", wrap_push_all_records)

    def build_pi_add_outgoing_connection(self, func: object):
        """
        Register a custom pi_add_outgoing_connection method.

        Parameters
        ----------
        func: Callable[object, str]
        The user-defined function that will be called by the Alteryx Engine, once for
        each defined output connection in the Plugin's Config.xml file.
        The function will return True to signify that the connection has been accepted.

        Returns
        -------
        None
        This method produces side-effects by registering the user defined function
        """

        @_monitor("pi_add_outgoing_connection")
        @wraps(func)
        def wrap_pi_add_outgoing_connection(current_plugin: object, str_name: str):
            try:
                return func(current_plugin, str_name)
            except Exception as e:
                logger.exception(e)
                raise e

        setattr(
            self._plugin, "pi_add_outgoing_connection", wrap_pi_add_outgoing_connection
        )

    def build_pi_close(self, func: object) -> None:
        """
        Register a custom pi_close method.

        Parameters
        ----------
        func: Callable[object, bool]
        The user-defined function that will be called by the Alteryx Engine, after
        all records for each of the defined incoming connections have been processed.

        Returns
        -------
        None
        This method produces side-effects by registering the user defined function
        """

        @_monitor("pi_close")
        @wraps(func)
        def wrap_pi_close(current_plugin: object, b_has_errors: bool) -> None:
            try:
                if current_plugin.all_inputs_completed:
                    func(current_plugin)
            except Exception as e:
                logger.exception(e)
                raise e

        setattr(self._plugin, "pi_close", wrap_pi_close)

    def build_ii_init(self, func: object):
        """
        Register a custom ii_init method.

        Parameters
        ----------
        func: Callable[object, object] -> bool
        The user-defined function that will be called by the Alteryx Engine to
        refresh metadata tracked by the plugin after changes to config or new tools are
        dragged onto the canvas.

        Returns
        -------
        None
        This method produces side-effects by registering the user defined function
        """

        @_monitor("ii_init")
        @wraps(func)
        def wrap_ii_init(current_interface: object, record_info_in: object):
            try:
                current_plugin = current_interface.parent
                current_plugin.update_sys_path()
                current_interface._interface_record_vars.record_info_in = record_info_in
                current_interface._interface_record_vars.fields = [
                    field for field in record_info_in
                ]
                current_interface._interface_record_vars.field_getters = {
                    field: interface_utils.get_getter_from_field(field)
                    for field in current_interface._interface_record_vars.fields
                }
                current_interface.initialized = True

                metadata = interface_utils.get_column_metadata(record_info_in)

                current_interface.anchor_metadata = metadata

                init_success = func(current_interface, record_info_in)

                if not init_success:
                    current_plugin.initialized = False
                    current_interface.initialized = False
                    return False

                if (
                    current_plugin.update_only_mode
                    and current_plugin.all_required_inputs_initialized
                ):
                    ret_val = self._init_func(current_plugin)
                    if ret_val:
                        self._build_metadata(current_plugin)
                        for (
                            _,
                            anchor,
                        ) in current_plugin._state_vars.output_anchors.items():
                            anchor.push_metadata(current_plugin)

                return True
            except Exception as e:
                logger.exception(e)
                raise e

        setattr(self._plugin.plugin_interface, "ii_init", wrap_ii_init)

    def build_ii_push_record(self, func: object):
        """
        Register a custom ii_push_record method.

        Parameters
        ----------
        func: Callable[object, object] -> bool
        The user-defined function that will be called by the Alteryx Engine, once
        for each incoming record for each input connection.

        Returns
        -------
        None
        This method produces side-effects by registering the user defined function
        """

        @_monitor("ii_push_record")
        @wraps(func)
        def wrap_ii_push_record(current_interface: object, in_record: sdk.RecordRef):
            try:
                current_plugin = current_interface.parent

                if not current_plugin.initialized or current_plugin.update_only_mode:
                    return False

                func(current_plugin, current_interface, in_record)
                return True
            except Exception as e:
                logger.exception(e)
                raise e

        setattr(self._plugin.plugin_interface, "ii_push_record", wrap_ii_push_record)

    def build_ii_update_progress(self, func: object):
        """
        Register a custom ii_update_progress method.

        Parameters
        ----------
        func: Callable[object, float]
        The user-defined function that will be called by the upstream tool,
        reporting the number of records it has pushed to the plugin.

        Returns
        -------
        None
        This method produces side-effects by registering the user defined function
        """

        @_monitor("ii_update_progress")
        @wraps(func)
        def wrap_ii_update_progress(current_interface: object, d_percentage: float):
            try:
                current_plugin = current_interface.parent
                if current_plugin.update_only_mode:
                    return

                current_plugin.update_progress(d_percentage)

                return func(current_interface, d_percentage)
            except Exception as e:
                logger.exception(e)
                raise e

        setattr(
            self._plugin.plugin_interface, "ii_update_progress", wrap_ii_update_progress
        )

    def build_ii_close(self, func: object):
        """
        Register a custom ii_close method.

        Parameters
        ----------
        func: Callable[object]
        The user-defined function that will be called by the Alteryx Engine,
        once for each incoming connection, after all records have been passed through
        ii_push_record step.

        Returns
        -------
        None
        This method produces side-effects by registering the user defined function
        """

        @_monitor("ii_close")
        @wraps(func)
        def wrap_ii_close(current_interface: object):
            try:
                current_plugin = current_interface.parent

                current_plugin.assert_all_inputs_connected()

                if current_plugin.update_only_mode:
                    return

                current_interface.completed = True

                return func(current_plugin)
            except Exception as e:
                logger.exception(e)
                raise e

        setattr(self._plugin.plugin_interface, "ii_close", wrap_ii_close)

    def initialize_plugin(self, func: object):
        """
        Decorate a user defined function to inject custom intialization.

        Parameters
        ----------
        func : Callable[[object, dict, object], bool]
        The user is expected to define a function which takes in
            User Defined Function Parameters (passed in at runtime):
            -------------------------------------------------------
                logger: An object which provides methods for logging to
                        Alteryx Designer results window
                workflow_config: A dictionary which is populated with the
                        values from the tool's configuration GUI, keys that
                        match the XML tags in the GUI.
                user_data: A SimpleNamespace object that allows the user to
                        store variables (for example values from the workflow_config)
                        to be accessed globally in later steps for the PluginFactory,
                        such as process_data.

            User Defined Function Returns
            -----------------------------
                bool
                    True if the variables needed to be passed  to tool for
                    data processing are present, False if otherwise.  This
                    manifests in Alteryx Designer as an error if the user
                    tries to run the tool but hasn't passed in proper values
                    to meet conditions that evaluate True.

        Returns
        -------
        This method doesn't return a value.  It creates side-effects by altering
        the state of the AyxPlugin class declaration that is returned by the
        PluginFactory.generate_plugin() method.

        Examples
        --------
            @factory.initialize_plugin
            def init(workflow_config, user_data, logger):
                user_data.some_variable = workflow_config.get(
                    "SomeVarInGuiXml")

                if user_data.some_variable is None:
                    logger.display_error_msg("User needs to input SomeVar")
                    init_success = False
                else:
                    init_success = True

                return init_success
        """

        @wraps(func)
        def wrap_init(current_plugin: object):
            current_plugin.initialized = _apply_parameter_requests(func)(current_plugin)
            return current_plugin.initialized

        self._init_func = wrap_init

    def build_metadata(self, func: object):
        """Decorate a function to inject user defined build metadata function."""

        def decorated_build_metadata(plugin: object):
            return _apply_parameter_requests(func)(plugin)

        self._build_metadata = decorated_build_metadata
        return

    def process_data(
        self, mode: str = "batch", input_type: str = "list", chunk_size: int = 1000
    ):
        """
        Decorate a function to inject user defined functionality.

        The process_data method is used to allow a user to inject
        Python code for processing records without having to specify
        boilerplate for input/output.

        The process data plugin method takes both direct parameters,
        mode and input_type, as well as indirect parameter of a user
        defined function, explained in greater detail below.

        Parameters
        ----------
        mode : str
        One of three options: 'batch', 'chunk', or 'stream'

        Batch mode will cause the tool to pull in all input records for
        a given input anchor, collect them into a single data
        structure (defined by input_type) and make this collection of records
        available to the User Defined Function in the respective input_anchor
        object.
        An important note for batch is that the User Defined Function will be
        executed one time, on the entire set of records at once.

        Chunk mode will cause the tool to pull a group (chunk) of records in
        at a time, and make this chunk of records available to the User
        Defined Function in the respective input_anchor object.
        In addition, a property on the interface object which will indicate
        whether or not the current chunk is the last chunk.

        Stream mode will cause the tool to pull one input record in at a time,
        and make this single record available to the User Defined Function in
        the respective input_anchor object.
        An important note for stream is that the User Defined Function will be
        executed once for every incoming record.

        input_type : str
        One of two options: 'list' or 'dataframe'
        Depending on the value set by user, the input data made available to
        the UDF by the respective input_anchor contained in the input_mgr
        will either contain records in the form of a Python list or a
        Pandas DataFrame object.

        chunk_size: int
        Determines the number of records which should make up each chunk. The
        final chunk will usually be less than the chunk_size, and this case
        is handled internally by this wrapper.

        Returns
        -------
        Callable[[Callable[object, object, Any], None]]
            The process_data method is a method which returns a
            higher order function.  This higher order function
            takes a user defined function as its parameter.

            The User Defined Function is a function defined by the
            user that will be called at runtime by the Alteryx Engine.
            It needs to meet specific requirements as to the parameters
            it accepts and the side effects it produces:

        Parameters
        ----------
        input_mgr : object
        The UDF will be given an input_mgr object as its first parameter.
        This object allows the user to reference any of a given plugin's
        defined input anchors, which facilitate fetching input data.

        output_mgr : object
        The UDF will be given an output_mgr object as its second parameter.
        This object allows the user to reference any of a given plugin's
        defined output anchors, which facilitate setting output data.

        user_data : object
        A SimpleNamespace object that allows the user to store variables
        (for example values from the workflow_config) to be accessed
        globally in later steps for the PluginFactory.  In the context of
        the UDF, this is a way to fetch values set in the initialize_plugin
        step to use for the process_data logic.

        Returns
        -------
        None
        The UDF doesn't return any values.  Instead, it needs to 'send'
        data as a side effect, by calling methods on the output_anchor
        object it gets from the output_mgr object it is provided.

        Examples
        --------
        @factory.process_data(mode="batch", input_type="dataframe")
        def process_data(input_mgr, output_mgr, user_data, logger):
            input_anchor = input_mgr.get_anchor("AnchorNameFromConfigXmlFile")
            input_df = input_anchor.get_data()

            output_df = input_df ## do stuff with data here

            output_anchor = output_mgr.get_anchor(
                "OutputAnchorNameFromConfigXmlFile")
            output_anchor.set_col_types(
                [sdk.FieldType.v_string, sdk.FieldType.int64])
            output_anchor.set_data(output_df)

        @factory.process_data(mode="chunk", input_type="dataframe", chunk_size=50)
        def process_data(input_mgr, output_mgr, user_data, logger):
            input_anchor = input_mgr.get_anchor("AnchorNameFromConfigXmlFile")
            input_df = input_anchor.get_data()

            if(input_anchor[0].is_last_chunk == False):
                message = f"This chunk contains records |{input_df.iloc[0]}| to |{input_df.iloc[49]}|"
                logger.display_info_msg(message)
            if(input_anchor[0].is_last_chunk):
                message = f"Last chunk contains records |{input_df.iloc[0]}| to |{input_df.iloc[49]}|"
                logger.display_info_msg(message)

        @factory.process_data(mode="stream")
        def process_each_record(input_mgr, output_mgr, user_data, logger):
            input_anchor = input_mgr.get_anchor("AnchorNameFromConfigXmlFile")
            input_row = input_anchor.get_data()
            col_names = input_anchor.get_col_names()
            col_types = input_anchor.get_col_types()

            output_row = []
            output_col_names = col_names + ["new_col_name"]
            output_col_types = col_types + [sdk.FieldType.v_string]

            for col in input_row:
                output_row.append(some_function(col))
            output_row.append("value for a new column")

            output_anchor = output_mgr.get_anchor(
                "OutputAnchorNameFromConfigXml")
            output_anchor.set_col_names(output_col_names)
            output_anchor.set_col_types(output_col_types)
            output_anchor.set_data(output_row)

        """
        # Save the requested data type for later
        setattr(self._plugin, "process_data_input_type", input_type)
        setattr(self._plugin, "process_data_mode", mode)

        def decorator_process_data(func: object):
            # Decorate user function to push all records and metadata
            func = _apply_parameter_requests(func)
            func = _push_all_metadata_and_records(func)

            @_run_only_if_pi_initialized
            def batch_ii_close(plugin: object):
                if plugin.all_inputs_completed:
                    # Run initialization here
                    if not self._init_func(plugin):
                        return

                    # Build metadata
                    self._build_metadata(plugin)

                    # Call user function to batch process data
                    func(plugin)

            @_run_only_if_pi_initialized
            def chunk_ii_push_record(
                plugin: object, current_interface: object, in_record: sdk.RecordRef
            ):
                current_interface.accumulate_record(in_record)

                if (
                    len(current_interface._interface_record_vars.record_list_in)
                    >= chunk_size
                ):

                    if current_interface.is_last_chunk == None:
                        plugin.initialized = self._init_func(plugin)
                        current_interface.is_last_chunk = False
                        if not plugin.initialized:
                            return
                        self._build_metadata(plugin)

                    func(plugin)

                    plugin.clear_accumulated_records()

            @_run_only_if_pi_initialized
            def chunk_ii_close(current_interface: object):
                plugin = current_interface.parent

                if plugin.update_only_mode:
                    return

                if current_interface.is_last_chunk == None:
                    plugin.initialized = self._init_func(plugin)
                    if not plugin.initialized:
                        return
                    self._build_metadata(plugin)

                if not plugin.initialized:
                    return

                current_interface.is_last_chunk = True

                func(plugin)

                plugin.clear_accumulated_records()

            @_run_only_if_pi_initialized
            def stream_ii_push_record(
                plugin: object, current_interface: object, in_record: sdk.RecordRef
            ):
                # Since we're streaming, we should clear any accumulated records
                plugin.clear_accumulated_records()

                # Then we can accumulate, this guarantees only one interface at a time
                # ever has a record
                current_interface.accumulate_record(in_record)

                self._build_metadata(plugin)

                func(plugin)

            @_run_only_if_pi_initialized
            def source_pi_push_all_records(plugin: object, n_record_limit: int):
                if not self._init_func(plugin):
                    return

                self._build_metadata(plugin)
                func(plugin)

            if mode.lower() == "batch":
                self.build_ii_push_record(
                    lambda plugin, interface, in_record: interface.accumulate_record(
                        in_record
                    )
                )
                self.build_ii_close(batch_ii_close)
            elif mode.lower() == "chunk":
                self.build_ii_push_record(chunk_ii_push_record)
                setattr(self._plugin.plugin_interface, "ii_close", chunk_ii_close)
            elif mode.lower() == "stream":
                # Streaming is the unique case for initialization
                # It should be ran in pi_init.
                self.build_pi_init(self._init_func)
                self.build_ii_push_record(stream_ii_push_record)
            elif mode.lower() == "source":
                self.build_pi_push_all_records(source_pi_push_all_records)
            else:
                err_str = """Mode parameter of process_data must be one of
                    'batch'|'chunk'|'stream'|'source'"""
                raise ValueError(err_str)

        return decorator_process_data

    def generate_plugin(self):
        """
        Return the constructed class definition for an AyxPlugin object.

        The AyxPlugin class definition is initialized by the Alteryx Engine via the
        Alteryx Python SDK.

        Parameters
        ----------
        None

        Returns
        -------
        object: AyxPlugin class definition
        The returned object is a modified AyxPlugin class definition, updated with
        the user definied functions injected by use of the various decorators the user
        has called.

        Example
        -------
            AyxPlugin = factory.generate_plugin()

        """
        return self._plugin


def _run_only_if_pi_initialized(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        plugin = args[0]
        if not plugin.initialized:
            return

        return func(*args, **kwargs)

    return wrapper


def _push_all_metadata_and_records(func):
    @wraps(func)
    def wrapper(plugin):
        func(plugin)

        plugin.push_all_metadata()
        plugin.push_all_output_records()

        if plugin.all_inputs_completed:
            plugin.close_all_outputs()

    return wrapper


def _apply_parameter_requests(func):
    @wraps(func)
    def wrapped(plugin):
        import inspect

        sig = inspect.signature(func)

        func_params = {
            "input_mgr": plugin.input_manager,
            "output_mgr": plugin.output_manager,
            "workflow_config": plugin.input_manager.workflow_config,
            "user_data": plugin.user_data,
            "logger": plugin.logging,
        }

        try:
            # Get the expected parameter names
            param_names = [name for name, param in sig.parameters.items()]
            passed_params = {name: func_params[name] for name in param_names}
        except KeyError:
            # Failed to build the requested params, using defaults
            return func(
                plugin.input_manager,
                plugin.output_manager,
                plugin.user_data,
                plugin.logging,
            )
        else:
            return func(**passed_params)

    return wrapped


def _monitor(name):
    def _monitor_decorator(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            if DEBUG:
                import time
                from pathlib import Path
                import pandas as pd

                start_time = time.time()
                val = func(*args, **kwargs)
                end_time = time.time()
                time_diff_ms = (end_time - start_time) * 1000

                # Write to a file
                file_name = Path("C:/debug.csv")

                if not file_name.is_file():
                    data = {"Func": [name], "Time": [time_diff_ms]}
                    df = pd.DataFrame(data)
                else:
                    data = {"Func": [name], "Time": [time_diff_ms]}
                    df_new = pd.DataFrame(data)
                    df_old = pd.read_csv(file_name)
                    df = pd.concat([df_old, df_new])

                df.to_csv(file_name, index=False)
                return val
            else:
                return func(*args, **kwargs)

        return wrapped

    return _monitor_decorator



#!/usr/bin/env python
from setuptools import setup, find_packages
from snakeplane.version import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="SnakePlane",
    url="https://github.com/alteryx/snakeplane",
    license="Apache License 2.0",
    version=__version__,
    python_requires=">=3.6.0",
    long_description=long_description,
    long_description_content_type="text/markdown",
    description="The Alteryx Python SDK Abstraction Layer",
    author="Alteryx",
    author_email="awalden@alteryx.com",
    install_requires=["numpy==1.16.3", "xmltodict==0.11.0"],
    packages=find_packages(),
)



# Copyright (C) 2019 Alteryx, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Snakeplane plugin utilities."""

# Built in Libraries
import os
import sys
from typing import Any, Dict, List


def split_abs_path(path: str) -> List[str]:
    """Split an absolute path into its parts, doing reduction on double dots."""
    allparts = []
    while 1:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        elif parts[1] == path:  # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            allparts.insert(0, parts[1])

    drop_idxs = []
    for idx, part in enumerate(allparts):
        if part == "..":
            drop_idxs.extend([idx - 1, idx])

    ret_val = []
    for idx, part in enumerate(allparts):
        if idx not in drop_idxs:
            ret_val.append(part)

    return ret_val


def contains_path(full: str, part: str) -> bool:
    """Return a boolean indicating if the part path is a subset of the full path."""
    full_parts = split_abs_path(full)
    part_parts = split_abs_path(part)

    for idx, name in enumerate(part_parts):
        if full_parts[idx] != name:
            return False

    return True


def get_tools_location() -> str:
    """Get the path to the Alteryx Python SDK Tools directory."""
    admin_path = os.path.join(os.environ["APPDATA"], "Alteryx", "Tools")
    user_path = os.path.join(os.environ["PROGRAMDATA"], "Alteryx", "Tools")
    alteryx_bin = (
        os.path.dirname(sys.executable)
        if "AlteryxEngineCmd.exe" in sys.executable
        else ""
    )
    html_plugins_path = os.path.join(alteryx_bin, "HtmlPlugins")
    if contains_path(__file__, admin_path):
        return admin_path

    if contains_path(__file__, user_path):
        return user_path

    if contains_path(__file__, alteryx_bin):
        return html_plugins_path

    raise RuntimeError("Tool is not located in Alteryx install locations.")


def get_tool_path(tool_name: str) -> str:
    """
    Generate the path to the installed location of the specified tool.

    Parameters
    ----------
    tool_name: str
        Name of the tool

    Returns
    -------
    str
        Absolute file path to the tool specified
    """
    return os.path.join(get_tools_location(), tool_name)


def get_xml_config_gui_settings(xml_dict: Dict[Any, Any]) -> Dict[Any, Any]:
    """
    Get the tool configuration from the config XML.

    Parameters
    ----------
    xml_dict: OrderedDictionary
        Parsed XML Tool configuration

    Returns
    -------
    OrderedDict
        GUI settings extracted from the parsed XML
    """
    return xml_dict["AlteryxJavaScriptPlugin"]["GuiSettings"]


def get_xml_config_input_connections(xml_dict: Dict[Any, Any]) -> List[Dict[Any, Any]]:
    """
    Get the input connection configuration from the tool XML.

    Parameters
    ----------
    xml_dict: OrderedDictionary
        Parsed XML Tool configuration

    Returns
    -------
    List[OrderedDict]
        List where each entry corresponds to an input anchor. Each entry
        is an ordered dictionary with anchor metadata.
    """
    connections = []
    inputs = get_xml_config_gui_settings(xml_dict).get("InputConnections")
    if inputs:
        connections = inputs.get("Connection")

    if connections and not isinstance(connections, List):
        connections = [connections]

    return connections


def get_xml_config_output_connections(xml_dict: Dict[Any, Any]) -> List[Dict[Any, Any]]:
    """
    Get the output connection configuration from the tool XML.

    Parameters
    ----------
    xml_dict: OrderedDictionary
        Parsed XML Tool configuration

    Returns
    -------
    List[OrderedDict]
        List where each entry corresponds to an output anchor. Each entry
        is an ordered dictionary with anchor metadata.
    """
    if "OutputConnections" not in get_xml_config_gui_settings(xml_dict):
        return []

    connections = get_xml_config_gui_settings(xml_dict)["OutputConnections"][
        "Connection"
    ]

    if not isinstance(connections, list):
        connections = [connections]

    return connections
